@@ -4,13 +4,13 @@ use 5.008000;
 use strict;
 use warnings;
 
-our $VERSION = '0.21_03';
+our $VERSION = '0.02';
 
 use File::Spec;
 use YAML::XS qw( LoadFile );
 use Cpanel::JSON::XS;
 use Hash::Merge;
-use Scalar::Util qw( refaddr readonly );
+use Data::Rmap qw( rmap_to :types );
 use Carp qw( croak );
 
 my %FILE_EXTENSIONS_MAP = (
@@ -43,26 +43,20 @@ Hash::Merge::specify_behavior(
 
 
 sub new {
-  my $class  = shift;
-  my %params = @_;
+  my $self_class = shift;
+  my %params     = @_;
 
-  my $self = bless {}, $class;
-
-  $self->{dirs} = $params{dirs} || [];
-  unless ( @{ $self->{dirs} } ) {
-    push( @{ $self->{dirs} }, '.' );
-  }
+  my $self = bless {}, $self_class;
 
+  $self->{dirs} = $params{dirs} || ['.'];
   $self->{interpolate_variables} = exists $params{interpolate_variables}
       ? $params{interpolate_variables} : 1;
   $self->{process_directives} = exists $params{process_directives}
       ? $params{process_directives} : 1;
-  $self->{export_env} = $params{export_env};
 
-  $self->{_merger} = Hash::Merge->new('CONFIG_PRECEDENT');
+  $self->{_hash_merge} = Hash::Merge->new( 'CONFIG_PRECEDENT' );
   $self->{_config}     = undef;
   $self->{_vars}       = {};
-  $self->{_seen_nodes} = {};
 
   return $self;
 }
@@ -70,13 +64,11 @@ sub new {
 {
   no strict 'refs';
 
-  foreach my $name (
-      qw( interpolate_variables process_directives export_env ) )
-  {
+  foreach my $name ( qw( interpolate_variables process_directives ) ) {
     *{$name} = sub {
       my $self = shift;
 
-      if (@_) {
+      if ( @_ ) {
         $self->{$name} = shift;
       }
 
@@ -86,24 +78,22 @@ sub new {
 }
 
 sub load {
-  my $self = shift;
+  my $self            = shift;
   my @config_sections = @_;
 
-  $self->{_config} = $self->_build_tree(@config_sections);
-  if ( $self->{export_env} ) {
-    $self->{_config} = $self->{_merger}->merge( $self->{_config},
-        { ENV => {%ENV} } );
-  }
-  $self->_process_tree( $self->{_config}, [] );
+  $self->{_config} = $self->_build_tree( @config_sections );
 
-  $self->{_vars}       = {};
-  $self->{_seen_nodes} = {};
+  rmap_to {
+    $_ = $self->_process_node( $_ );
+  } HASH|VALUE, $self->{_config};
+
+  $self->{_vars} = {};
 
   return $self->{_config};
 }
 
 sub _build_tree {
-  my $self = shift;
+  my $self            = shift;
   my @config_sections = @_;
 
   my $config = {};
@@ -112,7 +102,7 @@ sub _build_tree {
     next unless defined $config_section;
 
     if ( ref($config_section) eq 'HASH' ) {
-      $config = $self->{_merger}->merge( $config, $config_section );
+      $config = $self->{_hash_merge}->merge( $config, $config_section );
     }
     else {
       my %not_found_idx;
@@ -145,8 +135,8 @@ sub _build_tree {
               next;
             }
 
+            my $method = "_load_${file_type}_file";
             my @data = eval {
-              my $method = "_load_$file_type";
               return $self->$method($file_path);
             };
             if ($@) {
@@ -154,7 +144,7 @@ sub _build_tree {
             }
 
             foreach my $data_chunk (@data) {
-              $config = $self->{_merger}->merge( $config, $data_chunk );
+              $config = $self->{_hash_merge}->merge( $config, $data_chunk );
             }
           }
         }
@@ -174,14 +164,14 @@ sub _build_tree {
   return $config;
 }
 
-sub _load_yaml {
+sub _load_yaml_file {
   my $self      = shift;
   my $file_path = shift;
 
   return LoadFile($file_path);
 }
 
-sub _load_json {
+sub _load_json_file {
   my $self      = shift;
   my $file_path = shift;
 
@@ -192,48 +182,15 @@ sub _load_json {
   return @data;
 }
 
-sub _process_tree {
-  my $self = shift;
-  my $ancs = pop;
-
-  return if readonly( $_[0] );
-
-  $_[0] = $self->_process_node( $_[0], $ancs );
-
-  if ( my $node_addr = refaddr( $_[0] ) ) {
-    return if $self->{_seen_nodes}{$node_addr};
-
-    $self->{_seen_nodes}{$node_addr} = 1;
-  }
-
-  if ( ref( $_[0] ) eq 'HASH' ) {
-    foreach ( values %{ $_[0] } ) {
-      $self->_process_tree( $_, [ $_[0], @{$ancs} ] );
-    }
-  }
-  elsif ( ref( $_[0] ) eq 'ARRAY' ) {
-    foreach ( @{ $_[0] } ) {
-      $self->_process_tree( $_, [ $_[0], @{$ancs} ] );
-    }
-  }
-
-  return;
-}
-
 sub _process_node {
   my $self = shift;
   my $node = shift;
-  my $ancs = shift;
 
   return unless defined $node;
 
-  if ( !ref($node) && $self->{interpolate_variables} ) {
-    $node =~ s/\$((\$?)\{([^\}]*)\})/
-        $2 ? $1 : ( $self->_resolve_var( $3, [ @{$ancs} ] ) || '' )/ge;
-  }
-  elsif ( ref($node) eq 'HASH' && $self->{process_directives} ) {
+  if ( ref($node) eq 'HASH' && $self->{process_directives} ) {
     if ( defined $node->{var} ) {
-      $node = $self->_resolve_var( $node->{var}, [ @{$ancs} ] );
+      $node = $self->_resolve_var( $node->{var} );
     }
     elsif ( defined $node->{include} ) {
       $node = $self->_build_tree( $node->{include} );
@@ -241,17 +198,21 @@ sub _process_node {
     else {
       if ( defined $node->{underlay} ) {
         my $layer = delete $node->{underlay};
-        $layer = $self->_process_layer( $layer, $ancs );
-        $node = $self->{_merger}->merge( $layer, $node );
+        $layer = $self->_process_layer($layer);
+        $node = $self->{_hash_merge}->merge( $layer, $node );
       }
 
       if ( defined $node->{overlay} ) {
         my $layer = delete $node->{overlay};
-        $layer = $self->_process_layer( $layer, $ancs );
-        $node = $self->{_merger}->merge( $node, $layer );
+        $layer = $self->_process_layer($layer);
+        $node = $self->{_hash_merge}->merge( $node, $layer );
       }
     }
   }
+  elsif ( $self->{interpolate_variables} ) { # SCALAR
+    $node =~ s/\$((\$?)\{([^\}]*)\})/
+        $2 ? $1 : $self->_resolve_var( $3 )/ge;
+  }
 
   return $node;
 }
@@ -259,17 +220,16 @@ sub _process_node {
 sub _process_layer {
   my $self  = shift;
   my $layer = shift;
-  my $ancs  = shift;
 
   if ( ref($layer) eq 'HASH' ) {
-    $layer = $self->_process_node( $layer, $ancs );
+    $layer = $self->_process_node( $layer );
   }
   elsif ( ref($layer) eq 'ARRAY' ) {
     my $new_layer = {};
 
     foreach my $node ( @{$layer} ) {
-      $node = $self->_process_node( $node, $ancs );
-      $new_layer = $self->{_merger}->merge( $new_layer, $node );
+      $node = $self->_process_node( $node );
+      $new_layer = $self->{_hash_merge}->merge( $new_layer, $node );
     }
 
     $layer = $new_layer;
@@ -279,117 +239,61 @@ sub _process_layer {
 }
 
 sub _resolve_var {
-  my $self = shift;
-  my $name = shift;
-  my $ancs = shift;
+  my $self     = shift;
+  my $var_name = shift;
 
-  my $value;
+  my $vars = $self->{_vars};
 
-  if ( $name =~ m/^\./ ) {
-    my $node;
-    my @tokens = split( /\./, $name, -1 );
+  unless ( defined $vars->{$var_name}  ) {
+    my @tokens = split( /\./, $var_name );
+    my $pointer = $self->{_config};
 
-    while (1) {
-      my $token = $tokens[0];
-      $token =~ s/^\s+//;
-      $token =~ s/\s+$//;
+    while ( 1 ) {
+      my $token = shift @tokens;
 
-      last if length($token) > 0;
+      if ( ref($pointer) eq 'HASH' ) {
+        last unless defined $pointer->{$token};
 
-      shift @tokens;
+        if ( !@tokens ) {
+          $vars->{$var_name}
+              = $self->_process_node( $pointer->{$token} );
+          $pointer->{$token} = $vars->{$var_name};
 
-      last unless @tokens;
-      next unless @{$ancs};
+          last;
+        }
 
-      $node = shift @{$ancs};
-    }
+        last unless ref( $pointer->{$token} );
 
-    $value = eval {
-      $self->_fetch_value( $node, $ancs, \@tokens );
-    };
-
-    if ($@) {
-      chomp $@;
-      die qq{Can't resolve variable "$name"; $@\n};
-    }
-  }
-  else {
-    my $vars = $self->{_vars};
-
-    unless ( defined $vars->{$name} ) {
-      my @tokens = split( /\./, $name, -1 );
-
-      $vars->{$name} = eval {
-        $self->_fetch_value( $self->{_config}, [], \@tokens );
-      };
-
-      if ($@) {
-        chomp $@;
-        die qq{Can't resolve variable "$name"; $@\n};
+        $pointer = $pointer->{$token};
       }
-    }
+      else { # ARRAY
+        if ( $token =~ m/\D/ ) {
+          die "Argument \"$token\" isn't numeric in array element:"
+              . " $var_name\n";
+        }
 
-    $value = $vars->{$name};
-  }
+        last unless defined $pointer->[$token];
 
-  return $value;
-}
+        if ( !@tokens ) {
+          $vars->{$var_name}
+              = $self->_process_node( $pointer->[$token] );
+          $pointer->[$token] = $vars->{$var_name};
 
-####
-sub _fetch_value {
-  my $self   = shift;
-  my $node   = shift;
-  my $ancs   = shift;
-  my $tokens = shift;
+          last;
+        }
 
-  return $node unless @{$tokens};
+        last unless ref( $pointer->[$token] );
 
-  my $value;
-
-  while (1) {
-    my $token = shift @{$tokens};
-    $token =~ s/^\s+//;
-    $token =~ s/\s+$//;
-
-    if ( ref($node) eq 'HASH' ) {
-      last unless defined $node->{$token};
-
-      unshift( @{$ancs}, $node );
-
-      unless ( @{$tokens} ) {
-        $node->{$token} = $self->_process_node( $node->{$token}, $ancs );
-        $value = $node->{$token};
-
-        last;
+        $pointer = $pointer->[$token];
       }
-
-      last unless ref( $node->{$token} );
-
-      $node = $node->{$token};
-    }
-    else { # ARRAY
-      if ( $token =~ m/\D/ ) {
-        die qq{Argument "$token" isn't numeric in array element.\n};
-      }
-
-      last unless defined $node->[$token];
-
-      unshift( @{$ancs}, $node );
-
-      unless ( @{$tokens} ) {
-        $node->[$token] = $self->_process_node( $node->[$token], $ancs );
-        $value = $node->[$token];
-
-        last;
-      }
-
-      last unless ref( $node->[$token] );
-
-      $node = $node->[$token];
     }
   }
 
-  return $value;
+  unless ( defined $vars->{$var_name} ) {
+    $vars->{$var_name} = '';
+  }
+
+  return $vars->{$var_name};
 }
 
 1;
@@ -405,14 +309,10 @@ features
   use Config::Processor;
 
   my $config_processor = Config::Processor->new(
-    dirs => [qw( /etc/myapp /home/username/etc/myapp )]
+    dirs => [ qw( /etc/myapp /home/username/etc/myapp ) ],
   );
 
-  my $config = $config_processor->load(qw( dirs.yml db.json metrics/* ));
-
-  $config = $config_processor->load(
-    qw( dirs.yml db.json redis.yml mongodb.json metrics/* ),
-
+  my $config = $config_processor->load( qw( dirs.yml db.json metrics/* ),
     { myapp => {
         db => {
           connectors => {
@@ -439,14 +339,11 @@ F<.yaml>, F<.jsn>, F<.json>.
 =head2 new( %params )
 
   my $config_processor = Config::Processor->new(
-    dirs       => [qw( /etc/myapp /home/username/etc/myapp )],
-    export_env => 1,
+    dirs => [ qw( /etc/myapp /home/username/myapp/etc ) ],
   );
 
-  $config_processor = Config::Processor->new;
-
   $config_processor = Config::Processor->new(
-    dirs                  => [qw( /etc/myapp /home/username/etc/myapp )],
+    dirs                  => [ qw( /etc/myapp /home/username/myapp/etc ) ],
     interpolate_variables => 0,
     process_directives    => 0,
   );
@@ -455,26 +352,15 @@ F<.yaml>, F<.jsn>, F<.json>.
 
 =item dirs => \@dirs
 
-List of directories, in which configuration processor will search files. If
-the parameter not specified, current directory will be used.
+List of directories, in which configuration processor will search files.
 
 =item interpolate_variables => $boolean
 
-Enables or disables variable interpolation in configurations files.
-Enabled by default.
+Enables or disables variable interpolation. Enabled by default.
 
 =item process_directives => $boolean
 
-Enables or disables directive processing in configurations files.
-Enabled by default.
-
-=item export_env => $boolean
-
-Enables or disables environment variables exporting to configuration tree.
-If enabled, environment variables can be accessed by the key C<ENV> from the
-configuration tree and can be interpolated into other configuration parameters.
-
-Disabled by default.
+Enables or disables directive processing. Enabled by default.
 
 =back
 
@@ -482,6 +368,9 @@ Disabled by default.
 
 =head2 load( @config_sections )
 
+  my $config = $config_processor->load( qw( dirs.yml db.json metrics/* ),
+      \%local_config );
+
 Attempts to load all configuration sections and returns reference to resulting
 configuration tree.
 
@@ -489,8 +378,6 @@ Configuration section can be a relative filename, a filename with wildcard
 characters or a hash reference. Filenames with wildcard characters is processed
 by C<CORE::glob> function and supports the same syntax.
 
-  my $config = $config_processor->load( qw( myapp.yml extras/* ), \%hard_config );
-
 =head2 interpolate_variables( [ $boolean ] )
 
 Enables or disables variable interpolation in configurations files.
@@ -499,10 +386,6 @@ Enables or disables variable interpolation in configurations files.
 
 Enables or disables directive processing in configuration files.
 
-=head2 export_env( [ $boolean ] )
-
-Enables or disables environment variables exporting to configuration tree.
-
 =head1 MERGING RULES
 
 Config::Processor merges all configuration sections in one resulting configuration tree by following rules:
@@ -519,7 +402,7 @@ Config::Processor merges all configuration sections in one resulting configurati
 
   HASH \%a    SCALAR $b    SCALAR $b
   HASH \%a    ARRAY  \@b   ARRAY  \@b
-  HASH \%a    HASH   \%b   HASH   recursive_merge( \%a, \%b )
+  HASH \%a    HASH   \%b   HASH   { %a, %b }
 
 For example, we have two configuration files. F<db.yml> at the left side:
 
@@ -557,11 +440,8 @@ After merging of two files we will get:
 
 =head1 INTERPOLATION
 
-Config::Processor can interpolate variables in string values (if you need alias
-for complex structures see C<var> directive). Variable names can be absolute or
-relative. Relative variable names begins with dot "." The number of dots
-depends on the nesting level of the current configuration parameter relative to
-referenced configuration parameter.
+Config::Processor can interpolate variables in string values. For example, we
+have F<myapp.yml> file:
 
   myapp:
     media_formats: [ "images", "audio", "video" ]
@@ -569,11 +449,11 @@ referenced configuration parameter.
     dirs:
       root_dir: "/myapp"
       templates_dir: "${myapp.dirs.root_dir}/templates"
-      sessions_dir: "${.root_dir}/sessions"
+      sessions_dir: "${myapp.dirs.root_dir}/sessions"
       media_dirs:
-        - "${..root_dir}/media/${myapp.media_formats.0}"
-        - "${..root_dir}/media/${myapp.media_formats.1}"
-        - "${..root_dir}/media/${myapp.media_formats.2}"
+        - "${myapp.dirs.root_dir}/medis/${myapp.media_formats.0}"
+        - "${myapp.dirs.root_dir}/media/${myapp.media_formats.1}"
+        - "${myapp.dirs.root_dir}/media/${myapp.media_formats.2}"
 
 After processing of the file we will get:
 
@@ -607,14 +487,10 @@ After processing we will get:
 =item var: varname
 
 Assigns configuration parameter value to another configuration parameter.
-Variable names in the directive can be absolute or relative. Relative variable
-names begins with dot ".". The number of dots depends on the nesting level of
-the current configuration parameter relative to referenced configuration
-parameter.
 
   myapp:
     db:
-      default_options:
+      generic_options:
         PrintWarn:  0
         PrintError: 0
         RaiseError: 1
@@ -626,7 +502,7 @@ parameter.
           dbname:   "stat"
           username: "stat_writer"
           password: "stat_writer_pass"
-          options: { var: myapp.db.default_options }
+          options: { var: myapp.db.generic_options }
 
         stat_slave:
           host:     "stat-slave.mydb.com"
@@ -634,7 +510,7 @@ parameter.
           dbname:   "stat"
           username: "stat_reader"
           password: "stat_reader_pass"
-          options: { var: ...default_options }
+          options: { var: myapp.db.generic_options }
 
 =item include: filename
 
@@ -675,18 +551,18 @@ For example, you can use this directive to set default values of parameters.
             RaiseError: 1
 
         stat_master:
-          underlay: { var: .default }
+          underlay: { var: myapp.db.connectors.default }
           host:     "stat-master.mydb.com"
           username: "stat_writer"
           password: "stat_writer_pass"
 
         stat_slave:
-          underlay: { var: .default }
+          underlay: { var: myapp.db.connectors.default }
           host:     "stat-slave.mydb.com"
           username: "stat_reader"
           password: "stat_reader_pass"
 
-You can move default parameters in separate files.
+You can move default parameters in separate file.
 
   myapp:
     db:
@@ -696,19 +572,19 @@ You can move default parameters in separate files.
           - { include: db_connectors/default_test.yml }
 
         stat_master:
-          underlay: { var: .default }
+          underlay: { var: myapp.db.connectors.default }
           host:     "stat-master.mydb.com"
           username: "stat_writer"
           password: "stat_writer_pass"
 
         stat_slave:
-          underlay: { var: .default }
+          underlay: { var: myapp.db.connectors.default }
           host:     "stat-slave.mydb.com"
           username: "stat_reader"
           password: "stat_reader_pass"
 
         test:
-          underlay: { var: .default_test }
+          underlay: { var: myapp.db.connectors.default_test }
           username: "test"
           password: "test_pass"
 
@@ -738,18 +614,18 @@ configuration parameters.
           port: "4321"
 
         stat_master:
-          underlay: { var: .default }
+          underlay: { var: myapp.db.connectors.default }
           host:     "stat-master.mydb.com"
           username: "stat_writer"
           password: "stat_writer_pass"
-          overlay:  { var: .test }
+          overlay: { var: myapp.db.connectors.test }
 
         stat_slave:
-          underlay: { var: .default }
+          underlay: { var: myapp.db.connectors.default }
           host:     "stat-slave.mydb.com"
           username: "stat_reader"
           password: "stat_reader_pass"
-          overlay:  { var: .test }
+          overlay: { var: myapp.db.connectors.test }
 
 To disable overriding just assign to C<test> connector empty hash.
 
